"""
NMODL Asset Manager for Arbor and NEURON.

~ Glues your neurons together!

Manage a local NMODL library that's automatically compiled, loaded, and
recompiled whenever you change your NMODL code or simulator environment.
"""

from ._fs import get_cache_path as _get_cache_path
from ._glia import Glia, MechId
from .assets import Catalogue, Mod, Package
from .exceptions import (
    AssetLookupError,
    BuildCatalogueError,
    CatalogueError,
    CompileError,
    GliaError,
    LibraryError,
    ModSourceError,
    NeuronError,
    NmodlError,
    NoMatchesError,
    PackageApiError,
    PackageError,
    PackageFileError,
    PackageModError,
    PackageProjectError,
    PackageVersionError,
    ResolveError,
    TooManyMatchesError,
    UnknownAssetError,
)
from .neuron import MechAccessor

_manager = Glia()


def insert(section, asset, variant=None, pkg=None, /, attributes=None, x=None):
    """
    Insert a mechanism or point process into a Section.

    :param section: The section to insert the asset into.
    :type section: patch.objects.Section
    :param asset: The name of the asset. Will be resolved into a fully qualified
       NEURON name based on preferences, unless a fully qualified name is given.

    :type asset: str
    :param attributes: Attributes of the asset to set on the section/mechanism.
    :type attributes: dict
    :param pkg: Package preference. Overrides global & script preferences.
    :type pkg: str
    :param variant: Variant preference. Overrides global & script preferences.
    :type variant: str
    :param x: Position along the `section` to place the point process at. Does
       not apply to mechanisms.

    :type x: float
    :raises: LibraryError if the asset isn't found or was incorrectly marked as
       a point process.
    """
    return _manager.insert(section, asset, variant, pkg, attributes, x)


def resolve(asset, pkg=None, variant=None):
    """
    Resolve an asset selection command to the name known to NEURON.
    """
    return _manager.resolve(asset, pkg=pkg, variant=variant)


def select(asset, pkg=None, variant=None):
    """
    Set script scope preferences for an asset.

    :param asset: Unresolved asset name.
    :type asset: str
    :param pkg: Name of the package to prefer.
    :type pkg: str
    :param variant: Name of the variant to prefer.
    :type variant: str
    """
    return _manager.select(asset, pkg=pkg, variant=variant)


def compile():
    """
    Compile and test all mod files found in all Glia packages.
    """
    return _manager.compile()


def load_library():
    """
    Load the glia library (if it isn't loaded yet).
    """
    return _manager.load_library()


def context(assets=None, pkg=None, variant=None):
    """
    Creates a context that sets glia preferences during a `with` statement.
    """
    return _manager.context(assets=assets, pkg=pkg, variant=variant)


def catalogue(name):
    """
    Load or build an Arbor mechanism catalogue.

    :param name: Name of the Glia installed Arbor catalogue.
    :type name: str
    """
    return _manager.catalogue(name)


def package(name):
    """
    Return package.

    :param name: Name of the installed glia package.
    :type name: str
    """
    return _manager.package(name)


def get_packages():
    """
    Return all installed packages.
    """
    return [*_manager.packages]


def get_cache_path():
    """
    Get the cache path where intermediary files are stored.

    Glia also stores a JSON cache in another directory, which
    you won't find here, but ``glia cache --clear`` can reset.
    """
    return _get_cache_path()


__all__ = [
    "AssetLookupError",
    "BuildCatalogueError",
    "Catalogue",
    "CatalogueError",
    "CompileError",
    "GliaError",
    "LibraryError",
    "MechAccessor",
    "MechId",
    "Mod",
    "ModSourceError",
    "NeuronError",
    "NmodlError",
    "NoMatchesError",
    "Package",
    "PackageApiError",
    "PackageError",
    "PackageFileError",
    "PackageModError",
    "PackageProjectError",
    "PackageVersionError",
    "ResolveError",
    "TooManyMatchesError",
    "UnknownAssetError",
]



import contextlib
import hashlib
import typing
from pathlib import Path

if typing.TYPE_CHECKING:  # pragma: nocover
    from .assets import Package


def hash_update_from_file(filename, hash):
    msg = f"Trying to file-hash non file path '{filename}'"
    assert Path(filename).is_file(), msg
    with open(str(filename), "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash.update(chunk)
    return hash


def hash_file(filename):  # pragma: nocover
    return hash_update_from_file(filename, hashlib.md5()).hexdigest()


def hash_update_from_dir(directory, hash):
    msg = f"Trying to directory-hash non directory path '{directory}'"
    assert Path(directory).is_dir(), msg
    for path in sorted(Path(directory).iterdir()):
        hash.update(path.name.encode())
        if path.is_file():
            hash = hash_update_from_file(path, hash)
        elif path.is_dir():  # pragma: nocover
            hash = hash_update_from_dir(path, hash)
    return hash


def get_directory_hash(directory):
    return hash_update_from_dir(directory, hashlib.md5()).hexdigest()


def hash_path(path):
    return hashlib.md5(path.encode()).hexdigest()


def get_package_mods_hash(package: "Package"):
    h = hashlib.md5()
    for mod in package.mods:
        if not mod.path.exists():
            raise FileNotFoundError(
                f"Modfile {package.name}.{mod.mech_id} not found at '{mod.path}'"
            )
        hash_update_from_file(mod.path, h)
    return h.hexdigest()


def get_package_hash(package: "Package"):
    h = hashlib.md5()
    h.update(bytes(package.root))
    return h.hexdigest()


def hash_installs():
    hash = ""

    with contextlib.suppress(Exception):
        from patch.version import get_neuron_version

        hash += hash_path(str(get_neuron_version()))

    with contextlib.suppress(Exception):
        import arbor

        hash += hash_path(str(arbor.config()))

    return hash



import typing
import weakref

if typing.TYPE_CHECKING:  # pragma: nocover
    from .assets import Mod


class MechAccessor:
    def __init__(self, section, mod: "Mod", point_process=None):
        self._section_name = section.hname()
        self._section = weakref.proxy(section)
        self._mod = mod
        self._pp = point_process
        self._references = []

    def __neuron__(self):
        if self._pp is not None:
            try:
                from patch import transform

                return transform(self._pp)
            except Exception:
                pass
            return self._pp
        else:
            raise TypeError(
                "Density mechanisms can't be retrieved as a standalone Patch/NEURON "
                "entity. "
            )

    def __ref__(self, other):
        self._references.append(other)

    def __deref__(self, other):
        self._references.remove(other)

    @property
    def _connections(self):
        try:
            return self._pp._connections
        except AttributeError:
            raise TypeError(
                "Can't connect Patch/NEURON entities to a density mechanism."
            ) from None

    @_connections.setter
    def _connections(self, value):
        self._pp._connections = value

    def stimulate(self, *args, **kwargs):
        if self._pp is not None:
            return self._pp.stimulate(*args, **kwargs)
        else:
            raise TypeError("Can't stimulate a DensityMechanism.")

    def record(self, param=None, x=0.5):
        from patch import p

        if param is not None and self._pp is None:
            return p.record(
                getattr(self._section(x), f"_ref_{param}_{self._mod.mod_name}")
            )
        else:
            return p.record(self)

    def __record__(self):
        if self._pp is not None:
            return self._pp.__record__()
        else:
            raise TypeError(
                "No default record for DensityMechanisms, use .record('param') instead."
            )

    def set(self, attribute_or_dict, value=None, /, x=None):
        if value is None:
            for k, v in attribute_or_dict.items():
                self.set_parameter(k, v, x)
        else:
            self.set_parameter(attribute_or_dict, value, x)

    def set_parameter(self, param, value, x=None):
        mod = self._mod.mod_name
        if self._pp is not None:
            if param not in self._pp.parameters:
                raise AttributeError(
                    f"Point process {self._mod.mod_name} has no parameter '{param}'"
                )
            return setattr(self._pp, param, value)
        try:
            if x is None:
                setattr(self._section.__neuron__(), f"{param}_{mod}", value)
            else:
                setattr(getattr(self._section(x), mod), value)
        except ReferenceError:
            raise ReferenceError(
                "Trying to set attribute on section"
                f" '{self._section_name}' that has since been garbage collected"
            ) from None

    def get_parameter(self, param, x=None):
        if self._pp is not None:
            return getattr(self._pp, param)
        mod = self._mod.mod_name
        try:
            if x is None:
                return getattr(self._section.__neuron__(), f"{param}_{mod}")
            else:
                return getattr(getattr(self._section(x), mod), param)
        except ReferenceError:
            raise ReferenceError(
                "Trying to set attribute on section"
                f" '{self._section_name}' that has since been garbage collected"
            ) from None
        except AttributeError:
            raise AttributeError(
                f"Parameter '{param}' does not exist on {self._mod.asset_name}"
            ) from None

    @property
    def parameters(self):
        raise NotImplementedError(
            "Parameter overview not implemented yet. Use `get_parameter` instead."
        )



import contextlib
import os
from pathlib import Path

import toml
from black import find_project_root

from ..assets import Mod, SupportedDialect
from ..exceptions import PackageError, PackageFileError, PackageProjectError
from ._ast import NmodlWriter, PackageTransformer, get_package_transformer


class PackageManager:
    def __init__(self, path: str | os.PathLike):
        self._path = Path(path).resolve()
        self._pyproject = self._path / "pyproject.toml"
        if not self._pyproject.exists():
            raise PackageProjectError(
                f"'{self._path}' is not a Glia package root (no pyproject.toml)."
            )
        self._defaults = {"package": {"mod-dir": "mods"}}

    @classmethod
    def for_cwd(cls):
        root, method = find_project_root(tuple())
        return cls(root)

    @property
    def path(self):
        return self._path

    def load_pyproject(self):
        return toml.load(self._pyproject)

    def get_entry_point(self) -> str:
        entry_points = [
            *self.load_pyproject()
            .get("project", {})
            .get("entry-points", {})
            .get("glia.package", {})
            .values()
        ]
        if len(entry_points) != 1:
            raise PackageProjectError(
                "Each Glia project must advertise exactly one Glia package."
                ' Please add/modify the [project.entry-points."glia.package"] table in'
                f" '{self._pyproject}' ."
            )
        return entry_points[0]

    def get_module_path(self) -> Path:
        entry_point = self.get_entry_point()
        module_name, attr = _unpack_ep(entry_point)
        found_path = _try_find_project_module(self._path, module_name)
        if found_path is None:
            raise PackageFileError(
                f"Could not find Python file of advertised package '{entry_point}'."
            )
        if found_path.parent == self._path:
            raise PackageProjectError(
                f"Glia package '{found_path}' is directly inside project root."
                " Single file packages cannot bundle NMODL files."
            )
        return found_path

    def _get_transformer(self) -> PackageTransformer:
        entry_point = self.get_entry_point()
        module_name, attr = _unpack_ep(entry_point)
        path = self.get_module_path()
        attr = attr or "package"
        try:
            return get_package_transformer(path, attr)
        except PackageError as e:
            raise PackageError(
                "Could not determine find package declaration "
                f"from entry point '{entry_point}' in '{path}'."
            ) from e

    def get_mod_dir(self, mod_dir: str | os.PathLike = None):
        if mod_dir is None:
            mod_dir = self.get_setting("package", "mod-dir")
        mod_dir = Path(mod_dir)
        return self.get_package_path() / mod_dir

    def get_rel_path(self, mod_dir: str | os.PathLike = None):
        return self.get_mod_dir(mod_dir).relative_to(self.get_package_path())

    def get_package_path(self):
        return self.path / self.get_module_path().relative_to(self.path).parts[0]

    def get_mod_files(self, mod_dir: str | os.PathLike = None):
        return [*(self.get_mod_dir(mod_dir)).rglob("*.mod")]

    def get_mod_declarations(self):
        return self._get_transformer().get_modlist_declaration()

    def add_mod_file(self, source: Path, mod: "Mod"):
        transformer = self._get_transformer()
        mods = transformer.get_mods()
        if any(
            m.asset_name == mod.asset_name
            and m.variant == mod.variant
            and m.dialect == mod.dialect
            for m in mods
        ):
            raise ValueError("A mod with the same spec already exists")
        else:
            dest = self.get_package_path()
            writer = NmodlWriter(mod)
            writer.import_source(source, dest, dialect=mod.dialect)
            transformer.add_mod(mod)
            transformer.write_in_place()

    def get_setting(self, *settings: str):
        node = self.load_pyproject()
        for part in ("tool", "glia", *settings):
            node = node.get(part, None)
            if node is None:
                return self.get_default(*settings)
        return node

    def get_default(self, *settings: str):
        node = self._defaults
        for part in settings:
            node = node.get(part, None)
            if node is None:
                return None
        return node

    def get_package_shim(self):
        return self._get_transformer().get_package_shim()

    def get_mod_from_source(
        self,
        source: Path,
        name: str,
        variant: str = "0",
        dialect: SupportedDialect = None,
    ):
        mod = Mod(
            str(self.get_rel_path() / f"{name}__{variant}.mod"), name, variant=variant
        )
        mod.set_package(self.get_package_shim())
        writer = NmodlWriter(mod)
        writer.parse_source(source, dialect=dialect)
        writer.extract_source_info()
        return mod


def _try_find_project_module(path: Path, module_name) -> Path:
    module_root = path / module_name.replace(".", os.sep)
    module_path = module_root.with_suffix(".py")
    if module_path.exists():
        return module_path
    package_path = module_root / "__init__.py"
    if package_path.exists():
        return package_path


def _unpack_ep(ep):
    parts: list[str] = ep.split(":")
    attr = None
    with contextlib.suppress(IndexError):
        attr = parts[1]
    module_name = parts[0]
    return module_name, attr



from errr.tree import exception as _e
from errr.tree import make_tree as _make_tree

_make_tree(
    globals(),
    GliaError=_e(
        CompileError=_e(),
        LibraryError=_e(),
        NeuronError=_e(),
        ResolveError=_e(
            TooManyMatchesError=_e("matches", "asset", "pkg", "variant"),
            NoMatchesError=_e("pkg", "variant"),
            UnknownAssetError=_e(),
            AssetLookupError=_e(),
        ),
        PackageError=_e(
            PackageApiError=_e(),
            PackageFileError=_e(),
            PackageModError=_e(),
            PackageProjectError=_e(),
            PackageVersionError=_e(),
        ),
        CatalogueError=_e(BuildCatalogueError=_e()),
        NmodlError=_e(ModSourceError=_e()),
    ),
)

GliaError: type[Exception]
CompileError: type[GliaError]  # noqa: F821
LibraryError: type[GliaError]  # noqa: F821
NeuronError: type[GliaError]  # noqa: F821
ResolveError: type[GliaError]  # noqa: F821
TooManyMatchesError: type[ResolveError]  # noqa: F821
NoMatchesError: type[ResolveError]  # noqa: F821
UnknownAssetError: type[ResolveError]  # noqa: F821
AssetLookupError: type[ResolveError]  # noqa: F821
PackageError: type[GliaError]  # noqa: F821
PackageApiError: type[PackageError]  # noqa: F821
PackageFileError: type[PackageError]  # noqa: F821
PackageModError: type[PackageError]  # noqa: F821
PackageProjectError: type[PackageError]  # noqa: F821
PackageVersionError: type[PackageError]  # noqa: F821
CatalogueError: type[GliaError]  # noqa: F821
BuildCatalogueError: type[CatalogueError]  # noqa: F821
NmodlError: type[GliaError]  # noqa: F821
ModSourceError: type[NmodlError]  # noqa: F821


class PackageWarning(Warning):
    pass
